(* Code generation: translate takes a semantically checked AST and
produces LLVM IR

LLVM tutorial: Make sure to read the OCaml version of the tutorial

http://llvm.org/docs/tutorial/index.html

Detailed documentation on the OCaml LLVM library:

http://llvm.moe/
http://llvm.moe/ocaml/

*)

module L = Llvm
module A = Ast

module StringMap = Map.Make(String)

let translate (globals, functions) =
  let context = L.global_context () in
  let the_module = L.create_module context "Cryptal"
  and i32_t  = L.i32_type  context
  and i8_t   = L.i8_type   context in
  let i1_t   = L.i1_type   context
  and void_t = L.void_type context
  and str_t = L.pointer_type i8_t in
  let lat_pointer = L.pointer_type (L.i8_type context) in
  let gem_type = L.struct_type context  [| lat_pointer ; lat_pointer |] in
  let gem_pointer = L.pointer_type gem_type in

  let rec ltype_of_typ = function
      A.Int -> i32_t
    | A.Bool -> i1_t
    | A.Void -> void_t
    | A.Gem -> gem_type
    | A.UInt -> i32_t
    | A.Char -> i8_t
    | A.String -> str_t
    | A.Lattice -> lat_pointer
    | A.Pointer x -> L.pointer_type (ltype_of_typ x) in

  (* Declare each global variable; remember its value in a map *)
  let global_vars =
    let global_var m (t, n) =
      let init = L.const_int (ltype_of_typ t) 0
      in StringMap.add n ((L.define_global n init the_module), (t,0)) m in
    List.fold_left global_var StringMap.empty globals in

  (* Declare printf(), which the print built-in function will call *)
  let printf_t = L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
  let printf_func = L.declare_function "printf" printf_t the_module in

  let gem_add_t = L.function_type gem_type [| gem_pointer ; gem_pointer |] in
  let gem_add_func = L.declare_function "gem_add_func" gem_add_t the_module in

  let gem_sub_t = L.function_type gem_type [| gem_pointer ; gem_pointer |] in
  let gem_sub_func = L.declare_function "gem_sub_func" gem_sub_t the_module in

  let gem_mult_t = L.function_type gem_type [| gem_pointer ; gem_pointer |] in
  let gem_mult_func = L.declare_function "gem_mult_func" gem_mult_t the_module in

(* use lat pointers as a general 8 byte pointer *)
  let gem_mod_create_t = L.function_type lat_pointer [| i32_t ; i32_t |] in
  let gem_mod_create_func = L.declare_function "gem_mod_create_func" gem_mod_create_t the_module in

  let gem_int_create_t = L.function_type lat_pointer [| gem_pointer ; i32_t |] in
  let gem_int_create_func = L.declare_function "gem_int_create_func" gem_int_create_t the_module in

  let int_gem_create_t = L.function_type lat_pointer [|i32_t ; gem_pointer |] in
  let int_gem_create_func = L.declare_function "int_gem_create_func" int_gem_create_t the_module in

  let gem_mod_create_latint_t = L.function_type lat_pointer [| lat_pointer ; i32_t |] in
  let gem_mod_create_latint_func = L.declare_function "gem_mod_create_latint_func" gem_mod_create_latint_t the_module in

  let gem_mod_create_intlat_t = L.function_type lat_pointer [| i32_t ; lat_pointer |] in
  let gem_mod_create_intlat_func = L.declare_function "gem_mod_create_intlat_func" gem_mod_create_intlat_t the_module in

  let gem_mod_create_latlat_t = L.function_type lat_pointer [| lat_pointer ; lat_pointer |] in
  let gem_mod_create_latlat_func = L.declare_function "gem_mod_create_latlat_func" gem_mod_create_latlat_t the_module in

  let print_gem_t = L.function_type lat_pointer [|gem_type |] in
  let print_gem_func = L.declare_function "print_gem_func" print_gem_t the_module in

  let int_to_bn_t = L.function_type lat_pointer [| i32_t |] in
  let int_to_bn_func = L.declare_function "int_to_bn_func" int_to_bn_t the_module in
  (*We may not want to do modular division... *)
  let gem_div_t = L.function_type gem_type [| gem_pointer ; gem_pointer |] in
  let gem_div_func = L.declare_function "gem_div_func" gem_div_t the_module in

  let gem_pow_t = L.function_type gem_type [| gem_pointer ; gem_pointer |] in
  let gem_pow_func = L.declare_function "gem_pow_func" gem_pow_t the_module in

  let gem_eq_t = L.function_type i1_t [| gem_pointer ; gem_pointer |] in
  let gem_eq_func = L.declare_function "gem_eq_func" gem_eq_t the_module in

  let gem_neq_t = L.function_type i1_t [| gem_pointer ; gem_pointer |] in
  let gem_neq_func = L.declare_function "gem_neq_func" gem_neq_t the_module in

  let gem_le_t = L.function_type i1_t [|gem_pointer ; gem_pointer |] in
  let gem_le_func = L.declare_function "gem_le_func" gem_le_t the_module in

  let gem_leq_t = L.function_type i1_t [|gem_pointer ; gem_pointer |] in
  let gem_leq_func = L.declare_function "gem_leq_func" gem_leq_t the_module in

  let gem_ge_t = L.function_type i1_t [|gem_pointer ; gem_pointer |] in
  let gem_ge_func = L.declare_function "gem_ge_func" gem_ge_t the_module in

  let gem_geq_t = L.function_type i1_t [| gem_pointer ; gem_pointer |] in
  let gem_geq_func = L.declare_function "gem_geq_func" gem_geq_t the_module in

  let gem_inverse_t = L.function_type gem_type [| gem_pointer |] in
  let gem_inverse_func = L.declare_function "gem_inverse_func" gem_inverse_t the_module in


  (* int,gem ops *)
  let int_gem_add_t = L.function_type gem_type [| lat_pointer ; gem_pointer |] in
  let int_gem_add_func = L.declare_function "int_gem_add_func" int_gem_add_t the_module in

  let int_gem_sub_t = L.function_type gem_type [| lat_pointer ; gem_pointer |] in
  let int_gem_sub_func = L.declare_function "int_gem_sub_func" int_gem_sub_t the_module in

  let int_gem_mult_t = L.function_type gem_type [| lat_pointer ; gem_pointer |] in
  let int_gem_mult_func = L.declare_function "int_gem_mult_func" int_gem_mult_t the_module in

  let int_gem_div_t = L.function_type gem_type [| lat_pointer ; gem_pointer |] in
  let int_gem_div_func = L.declare_function "int_gem_div_func" int_gem_div_t the_module in

  let int_gem_pow_t = L.function_type gem_type [| lat_pointer ; gem_pointer |] in
  let int_gem_pow_func = L.declare_function "int_gem_pow_func" int_gem_pow_t the_module in

  (*int, gem comparisons *)
  let int_gem_eq_t = L.function_type i1_t [| lat_pointer ; gem_pointer |] in
  let int_gem_eq_func = L.declare_function "int_gem_eq_func" int_gem_eq_t the_module in

  let int_gem_neq_t = L.function_type i1_t [| lat_pointer ; gem_pointer |] in
  let int_gem_neq_func = L.declare_function "int_gem_neq_func" int_gem_neq_t the_module in

  let int_gem_less_t = L.function_type i1_t [| lat_pointer ; gem_pointer |] in
  let int_gem_less_func = L.declare_function "int_gem_less_func" int_gem_less_t the_module in

  let int_gem_greater_t = L.function_type i1_t [| lat_pointer ; gem_pointer |] in
  let int_gem_greater_func = L.declare_function "int_gem_greater_func" int_gem_greater_t the_module in

  let int_gem_leq_t = L.function_type i1_t [| lat_pointer ; gem_pointer |] in
  let int_gem_leq_func = L.declare_function "int_gem_leq_func" int_gem_leq_t the_module in

  let int_gem_geq_t = L.function_type i1_t [| lat_pointer ; gem_pointer |] in
  let int_gem_geq_func = L.declare_function "int_gem_geq_func" int_gem_geq_t the_module in

  let gem_get_t = L.function_type lat_pointer [| gem_type ; i32_t |] in
  let gem_get_func = L.declare_function "gem_get_func" gem_get_t the_module in

  let gem_to_lat_t = L.function_type lat_pointer [| gem_pointer |] in
  let gem_to_lat_func = L.declare_function "gem_to_lat_func" gem_to_lat_t the_module in

  (* Lattice ops *)
  let lat_str_create_t = L.function_type lat_pointer [| str_t |] in
  let lat_str_create_func = L.declare_function "lat_str_create_func" lat_str_create_t the_module in

  let print_lat_t = L.function_type lat_pointer [| lat_pointer |] in
  let print_lat_func = L.declare_function "print_lat_func" print_lat_t the_module in

  (* Lat, Lat ops *)
  let lat_add_t = L.function_type lat_pointer [| lat_pointer ; lat_pointer |] in
  let lat_add_func = L.declare_function "lat_add_func" lat_add_t the_module in

  let lat_sub_t = L.function_type lat_pointer [| lat_pointer ; lat_pointer |] in
  let lat_sub_func = L.declare_function "lat_sub_func" lat_sub_t the_module in

  let lat_mult_t = L.function_type lat_pointer [| lat_pointer ; lat_pointer |] in
  let lat_mult_func = L.declare_function "lat_mult_func" lat_mult_t the_module in

  let lat_div_t = L.function_type lat_pointer [| lat_pointer ; lat_pointer |] in
  let lat_div_func = L.declare_function "lat_div_func" lat_div_t the_module in

  let lat_pow_t = L.function_type lat_pointer [| lat_pointer ; lat_pointer |] in
  let lat_pow_func = L.declare_function "lat_pow_func" lat_pow_t the_module in

  let lat_mod_t = L.function_type lat_pointer [| lat_pointer ; lat_pointer |] in
  let lat_mod_func = L.declare_function "lat_mod_func" lat_mod_t the_module in

  let lat_eq_t = L.function_type i1_t [| lat_pointer ; lat_pointer |] in
  let lat_eq_func = L.declare_function "lat_eq_func" lat_eq_t the_module in

  let lat_neq_t = L.function_type i1_t [| lat_pointer ; lat_pointer |] in
  let lat_neq_func = L.declare_function "lat_neq_func" lat_neq_t the_module in

  let lat_le_t = L.function_type i1_t [|lat_pointer ; lat_pointer |] in
  let lat_le_func = L.declare_function "lat_le_func" lat_le_t the_module in

  let lat_leq_t = L.function_type i1_t [|lat_pointer ; lat_pointer |] in
  let lat_leq_func = L.declare_function "lat_leq_func" lat_leq_t the_module in

  let lat_ge_t = L.function_type i1_t [|lat_pointer ; lat_pointer |] in
  let lat_ge_func = L.declare_function "lat_ge_func" lat_ge_t the_module in

  let lat_geq_t = L.function_type i1_t [| lat_pointer ; lat_pointer |] in
  let lat_geq_func = L.declare_function "lat_geq_func" lat_geq_t the_module in

  let hash_md5_t = L.function_type str_t [|str_t|] in
  let hash_md5_func = L.declare_function "hash_md5_func" hash_md5_t the_module in 


  (* Declare the built-in printbig() function *)
  let printbig_t = L.function_type i32_t [| i32_t |] in
  let printbig_func = L.declare_function "printbig" printbig_t the_module in

  (* Declare the built-in print_string() function *)
  let print_string_t = L.function_type str_t [| str_t |] in
  let print_string_func = L.declare_function "print_string" print_string_t the_module in

  (* Define each function (arguments and return type) so we can call it *)
  let function_decls =
    let function_decl m fdecl =
      let name = fdecl.A.fname
      and formal_types =
    	Array.of_list (List.map (fun (t,_) -> ltype_of_typ t) fdecl.A.formals)
          in let ftype = L.function_type (ltype_of_typ fdecl.A.typ) formal_types in
          StringMap.add name (L.define_function name ftype the_module, fdecl) m in
        List.fold_left function_decl StringMap.empty functions in

  (* Fill in the body of the given function *)
  let build_function_body fdecl =
    let (the_function, _) = StringMap.find fdecl.A.fname function_decls in
    let builder = L.builder_at_end context (L.entry_block the_function) in
    let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
    and string_format_str = L.build_global_stringptr "%s\n" "fmt" builder in

    (* Construct the function's "locals": formal arguments and locally
     declared variables.  Allocate each on the stack, initialize their
     value, if appropriate, and remember their values in the "locals" map *)
     let lookup n tree = try StringMap.find n tree
     with Not_found -> StringMap.find n global_vars
     in


    (* Return the value for a variable or formal argument *)

    (* Construct code for an expression; return its value *)
    let rec expr  tree builder  = function
	      A.Literal i -> (L.const_int i32_t i, (A.Int,0))
      | A.BoolLit b -> (L.const_int i1_t (if b then 1 else 0), (A.Int, 0))
      | A.Noexpr -> (L.const_int i32_t 0, (A.Void, 0))
      | A.String s -> (L.build_global_stringptr s "fmt" builder, (A.Pointer(A.Char), 0))
      (* | A.Null ->  *)
      (* | A.Ch -> *)
      | A.Id s ->
      let binding = lookup s tree in
        (L.build_load (fst binding) s builder, (fst (snd binding), 1))

      (* Assign Value to Gem *)
      (* | A.Gem (e1, e2) ->
          (let newE1 = expr tree builder e1 and newE2 = expr tree buider e2 in
          let pointer  = L.build_alloca gem_type "e1e2" builder, (A.Gem, 0))


      *)
      (* | A.Lattice () *)
      | A.Gem (e1, e2) ->
        let (e1', (t1, _)) = expr tree builder e1
        and (e2', (t2, _)) = expr tree builder e2 in
        let struct_gem = L.undef gem_type in (match  (t1,t2) with
          (A.Int, A.Int) ->
            let modded_val = L.build_call gem_mod_create_func [|e1'; e2' |] "gem_mod_res" builder in
            let mod_val = L.build_call int_to_bn_func [|e2' |] "int_to_bn_res" builder in
            let struct_gem_final = L.build_insertvalue struct_gem (modded_val) 0 "gm" builder in
            (L.build_insertvalue struct_gem_final (mod_val) 1 "gm2" builder)
          | (A.Gem, A.Int) ->
            let pointer1 = L.build_alloca gem_type "e1" builder in
            let _ = L.build_store e1' pointer1 builder in
            let modded_val = L.build_call gem_int_create_func [| pointer1; e2' |] "gem_int_res" builder in
            let mod_val = L.build_call int_to_bn_func [| e2' |] "int_to_bn_res" builder in
            let struct_gem_final = L.build_insertvalue struct_gem (modded_val) 0 "gm" builder in
              (L.build_insertvalue struct_gem_final (mod_val) 1 "gm2" builder)
          | (A.Int, A.Gem) ->
            let pointer2 = L.build_alloca gem_type "e2" builder in
            let _ = L.build_store e2' pointer2 builder in
            let modded_val = L.build_call int_gem_create_func [| e1' ; pointer2 |] "int_gem_Res" builder in
            let mod_val = L.build_call int_to_bn_func [| e2' |] "int_to_bn_res" builder in
            let struct_gem_final = L.build_insertvalue struct_gem (modded_val) 0 "gm" builder in
              (L.build_insertvalue struct_gem_final (mod_val) 1 "gm2" builder)
          | (A.Int, A.Lattice) ->
            let modded_val = L.build_call gem_mod_create_intlat_func [| e1'; e2' |] "gem_mod_create_intlat_res" builder in
            let mod_val = e2' in
            let struct_gem_final = L.build_insertvalue struct_gem (modded_val) 0 "gm" builder in
            (L.build_insertvalue struct_gem_final (mod_val) 1 "gm2" builder)
          | (A.Lattice, A.Int) ->
            let modded_val = L.build_call gem_mod_create_latint_func [| e1'; e2' |] "gem_mod_create_latint_res" builder in
            let mod_val = L.build_call int_to_bn_func [|e2' |] "int_to_bn_res" builder in
            let struct_gem_final = L.build_insertvalue struct_gem (modded_val) 0 "gm" builder in
            (L.build_insertvalue struct_gem_final (mod_val) 1 "gm2" builder)
          | (A.Lattice, A.Lattice) ->
            let modded_val = L.build_call gem_mod_create_latlat_func [| e1'; e2' |] "gem_mod_create_latlat_res" builder in
            let mod_val = e2' in
            let struct_gem_final = L.build_insertvalue struct_gem (modded_val) 0 "gm" builder in
            (L.build_insertvalue struct_gem_final (mod_val) 1 "gm2" builder)
        ), (A.Gem, 1)
      (* Lattice literal - not used
      | A.Lat e ->
        let (e', (t, _)) = expr tree builder e in
        (match t with
            A.Pointer(A.Char) -> L.build_call lat_str_create_func [| e' |] "lat_str_res" builder
          | A.Int -> L.build_call int_to_bn_func [| e' |] "int_to_bn_res" builder
          | ty -> raise(Failure("cannot initialize lattice with " ^ A.string_of_typ ty))
        ), (A.Lattice, 1)
      *)
      | A.Binop (e1, op, e2) ->
      (* (print_string ("e1 is of type: " ^ (A.string_of_typ e1))); *)
    	  let (e1', (t1, leaf1)) = expr tree builder e1
        and  (e2', (t2, leaf2)) = expr tree builder e2 in
          (match (t1, t2) with
          | (A.Int, A.Int) ->
        	  ((match op with
          	    A.Add     -> L.build_add
          	  | A.Sub     -> L.build_sub
          	  | A.Mult    -> L.build_mul
              | A.Div     -> L.build_sdiv
          	  | A.And     -> L.build_and
          	  | A.Or      -> L.build_or
          	  | A.Equal   -> L.build_icmp L.Icmp.Eq
          	  | A.Neq     -> L.build_icmp L.Icmp.Ne
          	  | A.Less    -> L.build_icmp L.Icmp.Slt
          	  | A.Leq     -> L.build_icmp L.Icmp.Sle
          	  | A.Greater -> L.build_icmp L.Icmp.Sgt
          	  | A.Geq     -> L.build_icmp L.Icmp.Sge
              (* | A.Mod     -> L.  *)
              (* | A.Power ->  *)
        	  ) e1' e2' "tmp" builder, (A.Int, 0))
          | (A.Int, A.Gem) | (A.Lattice, A.Gem) ->
            let arg = if t1 = A.Int then (L.build_call int_to_bn_func [| e1' |] "int_to_bn_res1" builder) else e1' in
            let pointer1 = L.build_alloca gem_type "e2" builder in
            let _ = L.build_store e2' pointer1 builder in
            (match op with
              A.Add ->
                (L.build_call int_gem_add_func [|arg; pointer1 |] "gem_int_add_res" builder, (A.Gem, 0))
              | A.Sub ->
                (L.build_call int_gem_sub_func [|arg ; pointer1 |] "gem_int_sub_res" builder, (A.Gem, 0))
              | A.Mult ->
                (L.build_call int_gem_mult_func [|arg ; pointer1 |] "gem_int_mult_res" builder, (A.Gem, 0))
              | A.Div ->
                (L.build_call int_gem_div_func [|arg ; pointer1 |] "gem_int_div_res" builder, (A.Gem, 0))
              | A.Power ->
                (L.build_call int_gem_pow_func [|arg ; pointer1 |] "gem_int_pow_res" builder, (A.Gem, 0))
              | A.Equal ->
                (L.build_call int_gem_eq_func [|arg ; pointer1 |] "gem_int_add_res" builder, (A.Int, 0))
              | A.Neq ->
                (L.build_call int_gem_neq_func [|arg ; pointer1 |] "gem_int_neq_res" builder, (A.Int, 0))
              | A.Less ->
                (L.build_call int_gem_less_func [|arg ; pointer1 |] "gem_int_less_res" builder, (A.Int, 0))
              | A.Leq ->
                (L.build_call int_gem_leq_func [|arg ; pointer1 |] "gem_int_leq_res" builder, (A.Int, 0))
              | A.Greater ->
                (L.build_call int_gem_greater_func [|arg ; pointer1 |] "gem_int_greater_res" builder, (A.Int, 0))
              | A.Geq ->
                (L.build_call int_gem_geq_func [|arg ; pointer1 |] "gem_int_geq_res" builder, (A.Int, 0))
            )
          | (A.Gem, A.Int) | (A.Gem, A.Lattice) ->
            let arg = if t2 = A.Int then (L.build_call int_to_bn_func [| e2' |] "int_to_bn_res2" builder) else e2' in
            let pointer1 = L.build_alloca gem_type "e1" builder in
            let _ = L.build_store  e1' pointer1 builder in
            (match op with
              A.Add ->
                (L.build_call int_gem_add_func [|arg ; pointer1 |] "gem_int_add_res" builder, (A.Gem, 0))
              | A.Sub ->
                (L.build_call int_gem_sub_func [|arg ; pointer1 |] "gem_int_sub_res" builder, (A.Gem, 0))
              | A.Mult ->
                (L.build_call int_gem_mult_func [|arg ; pointer1 |] "gem_int_mult_res" builder, (A.Gem, 0))
              | A.Div ->
                (L.build_call int_gem_div_func [|arg ; pointer1 |] "gem_int_div_res" builder, (A.Gem, 0))
              | A.Power ->
                (L.build_call int_gem_pow_func [|arg ; pointer1 |] "gem_int_pow_res" builder, (A.Gem, 0))
              | A.Equal ->
                (L.build_call int_gem_eq_func [|arg ; pointer1 |] "gem_int_add_res" builder, (A.Int, 0))
              | A.Neq ->
                (L.build_call int_gem_neq_func [|arg ; pointer1 |] "gem_int_neq_res" builder, (A.Int, 0))
              | A.Less ->
                (L.build_call int_gem_less_func [|arg ; pointer1 |] "gem_int_less_res" builder, (A.Int, 0))
              | A.Leq ->
                (L.build_call int_gem_leq_func [|arg ; pointer1 |] "gem_int_leq_res" builder, (A.Int, 0))
              | A.Greater ->
                (L.build_call int_gem_greater_func [|arg ; pointer1 |] "gem_int_greater_res" builder, (A.Int, 0))
              | A.Geq ->
                (L.build_call int_gem_geq_func [|arg ; pointer1 |] "gem_int_geq_res" builder, (A.Int, 0))
              )

          | (A.Gem, A.Gem) ->
            let pointer1  = L.build_alloca gem_type "e1" builder and
            pointer2 = L.build_alloca gem_type "e2" builder in
            let _ = L.build_store e1' pointer1 builder and
            _ = L.build_store e2' pointer2 builder in
            (match op with
              A.Add ->
               (L.build_call gem_add_func [|pointer1 ; pointer2 |] "gem_add_res" builder, (A.Gem, 0))
              | A.Sub ->
                (L.build_call gem_sub_func [| pointer1 ; pointer2 |] "gem_sub_res" builder , (A.Gem, 0))
              | A.Mult ->
                (L.build_call gem_mult_func [| pointer1 ; pointer2 |] "gem_mult_res" builder, (A.Gem, 0))
              | A.Div ->
                (L.build_call gem_div_func [| pointer1 ; pointer2 |] "gem_div_res" builder, (A.Gem, 0))
              | A.Power ->
                (L.build_call gem_pow_func [| pointer1 ; pointer2 |] "gem_pow_res" builder, (A.Gem, 0))
              | A.Equal ->
                (L.build_call gem_eq_func [| pointer1 ; pointer2 |] "gem_eq_res" builder, (A.Int, 0))
              | A.Neq ->
               (L.build_call gem_neq_func [|pointer1 ; pointer2 |] "gem_neq_res" builder, (A.Int, 0))
              | A.Less ->
               (L.build_call gem_le_func [|pointer1 ; pointer2 |] "gem_le_res" builder, (A.Int,0))
              | A.Leq ->
               (L.build_call gem_leq_func [|pointer1 ; pointer2 |] "gem_leq_res" builder, (A.Int, 0))
              | A.Greater ->
               (L.build_call gem_ge_func [| pointer1 ; pointer2 |] "gem_ge_res" builder, (A.Int, 0))
              | A.Geq ->
               (L.build_call gem_geq_func [| pointer1 ; pointer2 |] "gem_geq_res" builder, (A.Int, 0))
              )

          | (A.Lattice, A.Lattice) | (A.Lattice, A.Int) | (A.Int, A.Lattice) ->
            (* convert the Int to a Lattice *)
            let arg1 = if t1 = A.Int then (L.build_call int_to_bn_func [| e1' |] "int_to_bn_res1" builder) else e1' in
            let arg2 = if t2 = A.Int then (L.build_call int_to_bn_func [| e2' |] "int_to_bn_res2" builder) else e2' in
            (match op with
                A.Add ->
                (L.build_call lat_add_func [| arg1 ; arg2 |] "lat_add_res" builder, (A.Lattice, 0))
              | A.Sub ->
                (L.build_call lat_sub_func [| arg1 ; arg2 |] "lat_sub_res" builder , (A.Lattice, 0))
              | A.Mult ->
                (L.build_call lat_mult_func [| arg1 ; arg2 |] "lat_mult_res" builder, (A.Lattice, 0))
              | A.Div ->
                (L.build_call lat_div_func [| arg1 ; arg2 |] "lat_div_res" builder, (A.Lattice, 0))
              | A.Power ->
                (L.build_call lat_pow_func [| arg1 ; arg2 |] "lat_pow_res" builder, (A.Lattice, 0))
              | A.Mod ->
                (L.build_call lat_mod_func [| arg1 ; arg2 |] "lat_mod_res" builder, (A.Lattice, 0))
              | A.Equal ->
                (L.build_call lat_eq_func [| arg1 ; arg2 |] "lat_eq_res" builder, (A.Int, 0))
              | A.Neq ->
               (L.build_call lat_neq_func [|arg1 ; arg2 |] "lat_neq_res" builder, (A.Int, 0))
              | A.Less ->
               (L.build_call lat_le_func [|arg1 ; arg2 |] "lat_le_res" builder, (A.Int,0))
              | A.Leq ->
               (L.build_call lat_leq_func [|arg1 ; arg2 |] "lat_leq_res" builder, (A.Int, 0))
              | A.Greater ->
               (L.build_call lat_ge_func [| arg1 ; arg2 |] "lat_ge_res" builder, (A.Int, 0))
              | A.Geq ->
               (L.build_call lat_geq_func [| arg1 ; arg2 |] "lat_geq_res" builder, (A.Int, 0))
              )

        )
      | A.Unop(op, e) ->
    	  let e', (t, _) = expr tree builder e in
    	  (( match t with
          A.Gem ->
            let pointer1 = L.build_alloca gem_type "e" builder in
            let _ = L.build_store e' pointer1 builder in
            (match op with
            A.Not -> L.build_call gem_inverse_func [| pointer1 |] "gem_inverse" builder )

          | _ ->(match op with
            A.Neg     -> L.build_neg e' "tmp" builder
          | A.Not     -> L.build_not e' "tmp" builder
        )) , (t, 0))
      | A.Assign (s, e) -> let (e', (t, _)) = expr tree builder e and
        left_type = (fst (snd (lookup s tree))) in (match (left_type, t) with
        (* Lattice Initialization cases *)
        | (A.Lattice, A.Pointer(A.Char)) ->
          let pointer =
            L.build_call lat_str_create_func [| e' |] "lat_str_res" builder in
          ignore(L.build_store pointer (fst (lookup s tree)) builder) ; (pointer, (left_type, 0))
        | (A.Lattice, A.Int) ->
          let pointer =
            L.build_call int_to_bn_func [| e' |] "int_to_bn_res" builder in
          ignore(L.build_store pointer (fst (lookup s tree)) builder) ; (pointer, (left_type, 0))
        | (A.Lattice, A.Gem) ->
          let pointer =
            let pointer1 = L.build_alloca gem_type "e" builder in
            let _ = L.build_store e' pointer1 builder in
            L.build_call gem_to_lat_func [| pointer1 |] "gem_to_lat_res" builder in
          ignore(L.build_store pointer (fst (lookup s tree)) builder) ; (pointer, (left_type, 0))
        (* Regular *)
        | _ -> ignore (L.build_store e' (fst (lookup s tree)) builder); ( e', (t, 0)) )

      (* | A.Call ("print", [e; expr_t]) | A.Call ("printb", [e; expr_t]) ->
        let var = expr builder in_b (e,expr_t) in
          (match expr_t with
            A.Int -> L.build_call printf_func [| int_format_str ; (var) |]
            | A.String -> L.build_call printf_func [| string_format_str ; (var)|]
            | _ -> raise(Failure("Couldn't match printf"))) "printf" builder *)
      (* | A.Call ("print", [e]) | A.Call ("printb", [e]) -> *)
      | A.Call ("gem_get", [e ; index ]) ->
       let (e', (t,_)) = expr tree builder e and
        (index', (t',_)) = expr tree builder index in
       (L.build_call gem_get_func [|e' ; index' |] "gem_get_func" builder, (A.Lattice, 0))
      | A.Call ("print_gem", [e] ) ->
      let (e', (t, _)) = expr tree builder e in
      (L.build_call print_gem_func [| e'|] "print_gem_func" builder, (t, 0))
      | A.Call ("print_lat", [e] ) ->
      let (e', (t, _)) = expr tree builder e in
      (L.build_call print_lat_func [| e'|] "print_lat_func" builder, (t, 0))
      | A.Call ("print", act) ->
        let actuals, _ = List.split (List.rev (List.map (expr tree builder)
        (List.rev act))) in
        let result = "0" in
        (L.build_call printf_func (Array.of_list actuals) result builder,
          (A.Pointer(A.Char), 0))
      (* | A.Call ("print_string", [e]) ->
        (* let actuals *)
        (L.build_call printf_func [| string_format_str ; (expr tree builder e)|] "print_string" builder,  (A.Pointer(A.Char), 0)) *)
      (* | A.Call ("printbig", [e]) ->
        L.build_call printbig_func [| (expr tree builder e) |] "printbig" builder, (A.Pointer(A.Char, 0)) *)
      | A.Call ("hash_md5", [e]) ->
        let (e', (t, _)) = expr tree builder e in
        (L.build_call hash_md5_func [| e'|] "hash_md5_func" builder, (t, 0))
      | A.Call (f, act) ->
        let (fdef, fdecl) = StringMap.find f function_decls in
        let actuals, _ = List.split (List.rev (List.map (expr tree builder) (List.rev act))) in
        let result =
          (match fdecl.A.typ with
              A.Void -> ""
            | _ -> f ^ "_result" ) in
        (L.build_call fdef (Array.of_list actuals) result builder, (fdecl.A.typ, 0))
      | e -> (print_string ((A.string_of_expr e) ^ " failed\n")) ;raise(Failure("illegal expression"))
    in

    (* Invoke "f builder" if the current block doesn't already
       have a terminal (e.g., a branch). *)
    let add_terminal builder f =
      match L.block_terminator (L.insertion_block builder) with
        Some _ -> ()
      | None -> ignore (f builder) in

    (* Build the code for the given statement; return the builder for
       the statement's successor *)
    let rec stmt tree builder = function
        A.Block (vl, sl) ->
          let new_tree =
            let add_local m (t,n) =
              let new_local = L.build_alloca (ltype_of_typ t) n builder
              in StringMap.add n (new_local, (t,0)) m
            in List.fold_left add_local tree vl
          in List.fold_left (stmt tree) builder sl

      | A.Expr e -> ignore (expr tree builder e); builder
      | A.Return e -> ignore (match fdecl.A.typ with
          A.Void -> L.build_ret_void builder
        | _ -> L.build_ret (fst (expr tree builder e)) builder); builder
      | A.If (predicate, then_stmt, else_stmt) ->
        let bool_val = fst( expr tree builder predicate) in
        let merge_bb = L.append_block context "merge" the_function in

        let then_bb = L.append_block context "then" the_function in
        add_terminal (stmt tree (L.builder_at_end context then_bb) then_stmt)
         (L.build_br merge_bb);

        let else_bb = L.append_block context "else" the_function in
        add_terminal (stmt tree (L.builder_at_end context else_bb) else_stmt)
         (L.build_br merge_bb);

        ignore (L.build_cond_br bool_val then_bb else_bb builder);
        L.builder_at_end context merge_bb

      | A.While (predicate, body) ->
        let pred_bb = L.append_block context "while" the_function in
        ignore (L.build_br pred_bb builder);

        let body_bb = L.append_block context "while_body" the_function in
        add_terminal (stmt tree (L.builder_at_end context body_bb) body)
          (L.build_br pred_bb);

        let pred_builder = L.builder_at_end context pred_bb in
        let bool_val = fst( expr tree pred_builder predicate) in

        let merge_bb = L.append_block context "merge" the_function in
        ignore (L.build_cond_br bool_val body_bb merge_bb pred_builder);
        L.builder_at_end context merge_bb

      | A.For (e1, e2, e3, body) -> stmt tree builder
        ( A.Block ([], [A.Expr e1 ; A.While (e2, A.Block ([], [body ; A.Expr e3])) ] ))
    in
    let local_vars =
      let add_formal m (t, n) p = L.set_value_name n p;
        let local = L.build_alloca (ltype_of_typ t) n builder in
        ignore (L.build_store p local builder);
        StringMap.add n (local, (t,0)) m in

      let add_local m (t, n) =
        let local_var = L.build_alloca (ltype_of_typ t) n builder
        in StringMap.add n (local_var, (t,0)) m in

      let formals = List.fold_left2 add_formal StringMap.empty fdecl.A.formals
          (Array.to_list (L.params the_function)) in
      List.fold_left add_local formals fdecl.A.locals in

    (* Build the code for each statement in the function *)
    let builder = stmt local_vars builder (A.Block ([], fdecl.A.body)) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder (match fdecl.A.typ with
        A.Void -> L.build_ret_void
      | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
  in

  List.iter build_function_body functions;
  the_module
